#!/usr/bin/env python
# (C) Copyright 2021 JSC.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, JSC does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import xarray as xr
from typing import Union, List

List_or_tuple = Union[List, tuple]

class Processing_Unet_Data(object):
    """
    Class to preprocess data for U-net model
    """
    class_name = "Processing_Unet_Data"

    def __init__(self, data, hold_data=True, dims: List[str] = ()):
        """
        Initialize processing instance
        :param data: data used for retrieving initial normalization parameters
        :param hold_data: flag if data is hold as attribute of the class instance
        :param dims: list of dimension over which parameters are derived
        """
        method = "{0}->{1}".format(Processing_Unet_Data.class_name, Processing_Unet_Data.__init__.__name__)

        # sanity checks
        assert isinstance(data, xr.DataArray), "%{0}: data must be a xarray DataArray.".format(method)
        assert isinstance(dims, List_or_tuple), "%{0}: dims must be a list or tuple of named dimensions.".format(method)

        self.norm_name = "z_score"
        self.mu, self.std = Processing_Unet_Data.get_z_norm_stat(data, dims)
        self.dims = dims
        self.data = None
        if hold_data:
            self.data = data

    @staticmethod
    def get_z_norm_stat(data: xr.DataArray, dims: List[str] = ()):
        """
        Retrieve parameters for z-score normalization.
        :param data: the data from which normalization-parameters are retrieved
        :param dims: list of dimension over which parameters are derived
        :return: mean and standard deviation
        """
        method = "{0}->{1}".format(Processing_Unet_Data.class_name, Processing_Unet_Data.get_z_norm_stat.__name__)

        # sanity checks
        assert isinstance(data, xr.DataArray), "%{0}: data must be a xarray DataArray.".format(method)
        assert isinstance(dims, List_or_tuple), "%{0}: dims must be a list or tuple of named dimensions.".format(method)

        if not dims:
            dims = list(data.dims)
        try:
            mu = data.mean(dim=dims)
            std = data.std(dim=dims)
        except Exception as err:
            print("%{0}: Failed to calculate mean and standard deviation for normalization. See risen error"
                  .format(method))
            raise err

        return mu, std

    def z_norm_data(self, data: xr.DataArray = None, reuse_stat: bool = True, dims: List[str] = (),
                    return_stat: bool = False):
        """
        Perform z-score normalization on data array. The average (mu) and standard deviation can be predescribed or
        calculated based on the data
        :param data: the data array to be normalized
        :param reuse_stat: flag if statistics from init-method is re-used
        :param dims: list of dimension onto which to operate if mu and std are calculated
        :param return_stat: flag if mu and std are returned
        :return: normalized data and optionally, mu and std
        """
        method = "{0}->{1}".format(Processing_Unet_Data.class_name, Processing_Unet_Data.z_norm_data.__name__)

        if data is None:
            if self.data is None:
                raise ValueError("%{0}: data-attribute is None and no data was passed.".format(method))
            else:
                data = self.data

        assert isinstance(data, xr.DataArray), "%{0}: data must be a xarray DataArray.".format(method)

        if reuse_stat:
            if dims:
                assert dims == self.dims, "Dimension over which statistics are calculated do not match.".format(method)
            mu, std = self.mu, self.std
        else:
            mu, std = Processing_Unet_Data.get_z_norm_stat(data, dims)

        if return_stat:
            return (data - mu) / std, mu, std
        else:
            return (data - mu) / std

